import pytest
import io
import tempfile
import os
import subprocess


def test_cksum_import():
    """Test that the cksum module can be imported."""
    from NASADEM.cksum import cksum


def test_cksum_with_bytes():
    """Test cksum function with bytes input."""
    from NASADEM.cksum import cksum
    
    test_data = b'Hello World'
    result = cksum(test_data)
    
    # This should be a positive integer
    assert isinstance(result, int)
    assert result > 0
    
    # Known checksum for "Hello World"
    assert result == 3576645817


def test_cksum_with_file_object():
    """Test cksum function with file-like object."""
    from NASADEM.cksum import cksum
    
    test_data = b'Hello World'
    file_obj = io.BytesIO(test_data)
    result = cksum(file_obj)
    
    assert isinstance(result, int)
    assert result == 3576645817


def test_cksum_matches_system_cksum():
    """Test that our cksum implementation matches system cksum command."""
    from NASADEM.cksum import cksum
    
    # Create a temporary file with known content
    test_data = b'Test data for checksum verification'
    
    with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
        tmp_file.write(test_data)
        tmp_file.flush()
        
        try:
            # Get system cksum result
            result = subprocess.run(['cksum', tmp_file.name], 
                                  capture_output=True, text=True)
            if result.returncode == 0:
                system_cksum = int(result.stdout.split()[0])
                
                # Get our implementation result
                with open(tmp_file.name, 'rb') as f:
                    our_cksum = cksum(f)
                
                assert our_cksum == system_cksum
            else:
                pytest.skip("System cksum command not available")
        finally:
            os.unlink(tmp_file.name)


def test_cksum_empty_data():
    """Test cksum with empty data."""
    from NASADEM.cksum import cksum
    
    result = cksum(b'')
    assert isinstance(result, int)
    # Empty file should have a specific checksum
    assert result == 4294967295  # 0xFFFFFFFF for empty input


import pytest

# List of dependencies
dependencies = [
    "colored_logging",
    "rasters",
    "xmltodict"
]

# Generate individual test functions for each dependency
@pytest.mark.parametrize("dependency", dependencies)
def test_dependency_import(dependency):
    __import__(dependency)



"""
Pure Python implementation of POSIX cksum algorithm.

This module provides a replacement for the pycksum package that is compatible
with Python 3.12 and later versions.
"""


def cksum(data_or_file):
    """
    Calculate POSIX cksum checksum for data or file-like object.
    
    Args:
        data_or_file: Either bytes data or a file-like object opened in binary mode
        
    Returns:
        int: The POSIX cksum checksum value
    """
    # Handle file-like objects
    if hasattr(data_or_file, 'read'):
        data = data_or_file.read()
    else:
        data = data_or_file
    
    # Ensure we have bytes
    if isinstance(data, str):
        data = data.encode('utf-8')
    
    # Initialize CRC with 0
    crc = 0
    
    # Process each byte of data
    for byte in data:
        # XOR the byte with the current CRC (shifted left 8 bits) 
        crc ^= byte << 24
        
        # Process 8 bits
        for _ in range(8):
            if crc & 0x80000000:  # If MSB is set
                crc = (crc << 1) ^ 0x04c11db7  # CRC-32 polynomial
            else:
                crc = crc << 1
            crc &= 0xffffffff  # Keep it 32-bit
    
    # Append the length in bytes as a big-endian value
    length = len(data)
    length_bytes = []
    while length > 0:
        length_bytes.insert(0, length & 0xff)
        length >>= 8
    
    # Process the length bytes
    for byte in length_bytes:
        crc ^= byte << 24
        for _ in range(8):
            if crc & 0x80000000:
                crc = (crc << 1) ^ 0x04c11db7
            else:
                crc = crc << 1
            crc &= 0xffffffff
    
    # Final XOR and return as signed 32-bit integer equivalent
    result = crc ^ 0xffffffff
    
    # Convert to match expected return type (unsigned 32-bit integer)
    return result


import netrc
import hashlib
import logging
import os
import posixpath
import re
import shutil
import urllib
from datetime import date
from fnmatch import fnmatch
from http.cookiejar import CookieJar
from os import makedirs, remove
from os.path import dirname
from os.path import exists
from os.path import getsize
from os.path import isdir
from os.path import join
from os.path import abspath
from os.path import expanduser
from time import sleep
from typing import List, OrderedDict

import requests
import xmltodict
from bs4 import BeautifulSoup
from dateutil import parser
from .cksum import cksum

import colored_logging as cl


class DownloadFailed(Exception):
    pass

CONNECTION_CLOSE = {
    "Connection": "close",
}

DEFAULT_REMOTE = "https://e4ftl01.cr.usgs.gov"
RETRIES = 6
WAIT_SECONDS = 360
XML_RETRIES = RETRIES
XML_TIMEOUT = WAIT_SECONDS
DOWNLOAD_RETRIES = RETRIES
DOWNLOAD_WAIT_SECONDS = WAIT_SECONDS

__author__ = "Gregory Halverson"

logger = logging.getLogger(__name__)


class LPDAACServerUnreachable(ConnectionError):
    pass


class LPDAACDataPool:
    logger = logging.getLogger(__name__)
    DEFAULT_CHUNK_SIZE = 2 ** 20
    DATE_REGEX = re.compile(r'^(19|20)\d\d[- /.](0[1-9]|1[012])[- /.](0[1-9]|[12][0-9]|3[01])$')
    DEFAULT_REMOTE = DEFAULT_REMOTE

    def __init__(self, username: str = None, password: str = None, remote: str = None, offline_ok: bool = True):
        if remote is None:
            remote = DEFAULT_REMOTE

        if username is None or password is None:
            try:
                netrc_file = netrc.netrc()
                username, _, password = netrc_file.authenticators("urs.earthdata.nasa.gov")
            except Exception as e:
                logger.warning("netrc credentials not found for urs.earthdata.nasa.gov")

        if username is None or password is None:
            if not "LPDAAC_USERNAME" in os.environ or not "LPDAAC_PASSWORD" in os.environ:
                logger.warning("missing environment variable 'LPDAAC_USERNAME' or 'LPDAAC_PASSWORD'")
            else:
                username = os.environ["LPDAAC_USERNAME"]
                password = os.environ["LPDAAC_PASSWORD"]

        self._remote = remote
        self._username = username
        self._password = password
        self.offline_ok = offline_ok

        # if self.offline_ok:
        #     logger.warning("going into offline mode")

        self._listings = {}

        if not self.offline_ok:
            try:
                self._authenticate()
                self._check_remote()
            except Exception as e:
                if self.offline_ok:
                    logger.warning("unable to connect to LP-DAAC data pool")
                else:
                    raise e

    def _authenticate(self):
        try:
            # https://wiki.earthdata.nasa.gov/display/EL/How+To+Access+Data+With+Python

            password_manager = urllib.request.HTTPPasswordMgrWithDefaultRealm()

            password_manager.add_password(
                realm=None,
                uri="https://urs.earthdata.nasa.gov",
                user=self._username,
                passwd=self._password
            )

            cookie_jar = CookieJar()

            # Install all the handlers.

            opener = urllib.request.build_opener(
                urllib.request.HTTPBasicAuthHandler(password_manager),
                # urllib2.HTTPHandler(debuglevel=1),    # Uncomment these two lines to see
                # urllib2.HTTPSHandler(debuglevel=1),   # details of the requests/responses
                urllib.request.HTTPCookieProcessor(cookie_jar)
            )

            urllib.request.install_opener(opener)
        except Exception as e:
            message = "unable to authenticate with LP-DAAC data pool"
            if self.offline_ok:
                logger.warning(message)
            else:
                raise ConnectionError(message)

    def _check_remote(self):
        logger.debug(f"checking URL: {cl.URL(self.remote)}")

        try:
            response = requests.head(self.remote, headers=CONNECTION_CLOSE)
            status = response.status_code
            duration = response.elapsed.total_seconds()
        except Exception as e:
            logger.exception(e)
            message = f"unable to connect to URL: {self.remote}"

            if self.offline_ok:
                logger.warning(message)
                return
            else:
                raise LPDAACServerUnreachable(message)

        if status == 200:
            logger.debug(
                "remote verified with status " + cl.val(200) +
                " in " + cl.time(f"{duration:0.2f}") +
                " seconds: " + cl.URL(self.remote))
        else:
            message = f"status: {status} URL: {self.remote}"

            if self.offline_ok:
                logger.warning(message)
            else:
                raise ConnectionError(message)

    @property
    def remote(self):
        return self._remote

    def get_HTTP_text(self, URL: str) -> str:
        try:
            request = urllib.request.Request(URL)
            response = urllib.request.urlopen(request)
            body = response.read().decode()
        except Exception as e:
            logger.exception(e)
            raise ConnectionError(f"cannot connect to URL: {URL}")

        return body

    def get_HTTP_listing(self, URL: str, pattern: str = None) -> List[str]:
        if URL in self._listings:
            listing = self._listings[URL]
        else:
            text = self.get_HTTP_text(URL)
            soup = BeautifulSoup(text, 'html.parser')
            links = list(soup.find_all('a', href=True))

            # get directory names from links on http site
            listing = sorted([link['href'].replace('/', '') for link in links])
            self._listings[URL] = listing

        if pattern is not None:
            listing = sorted([
                item
                for item
                in listing
                if fnmatch(item, pattern)
            ])

        return listing

    def get_HTTP_date_listing(self, URL: str) -> List[date]:
        return sorted([
            parser.parse(item).date()
            for item
            in self.get_HTTP_listing(URL)
            if self.DATE_REGEX.match(item)
        ])

    def read_HTTP_XML(self, URL: str) -> OrderedDict:
        return xmltodict.parse(self.get_HTTP_text(URL))

    def generate_XML_URL(self, URL: str) -> str:
        return f"{URL}.xml"

    def get_metadata(self, data_URL: str) -> OrderedDict:
        metadata_URL = f"{data_URL}.xml"
        logger.info(f"checking metadata: {cl.URL(metadata_URL)}")
        request = urllib.request.Request(metadata_URL)
        response = urllib.request.urlopen(request)
        duration = response.elapsed.total_seconds()
        body = response.read().decode()
        metadata = xmltodict.parse(body)
        logger.info(f"metadata retrieved in {cl.val(f'{duration:0.2f}')} seconds: {cl.URL(metadata_URL)}")

        return metadata

    def get_remote_checksum(self, URL: str) -> int:
        return int(self.get_metadata(URL)["GranuleMetaDataFile"]["GranuleURMetaData"]["DataFiles"]["DataFileContainer"][
                       "Checksum"])

    def get_remote_filesize(self, URL: str) -> int:
        return int(self.get_metadata(URL)["GranuleMetaDataFile"]["GranuleURMetaData"]["DataFiles"]["DataFileContainer"][
                       "FileSize"])

    def get_local_checksum(self, filename: str, checksum_type: str = "CKSUM") -> str:
        with open(filename, "rb") as file:
            if checksum_type == "CKSUM":
                return str(int(cksum(file)))
            elif checksum_type == "MD5":
                return str(hashlib.md5(file.read()).hexdigest())

    def get_local_filesize(self, filename: str) -> int:
        return getsize(filename)

    def product_directory(self, platform: str, product: str, build: str = None) -> str:
        if build is None:
            build = "001"
        elif isinstance(build, float):
            build = f"{int(build * 10)}:03d"
        elif isinstance(build, int):
            build = f"{build:03d}"

        URL = posixpath.join(self._remote, platform, f"{product}.{build}")

        return URL

    def dates(self, platform: str, product: str, build: str = None) -> List[date]:
        return self.get_HTTP_date_listing(self.product_directory(platform, product, build))

    def date_URL(
            self,
            platform: str,
            product: str,
            acquisition_date: date or str,
            build: str = None) -> str:
        if isinstance(acquisition_date, str):
            acquisition_date = parser.parse(acquisition_date).date()

        URL = posixpath.join(
            self.product_directory(platform, product, build),
            f"{acquisition_date:%Y.%m.%d}"
        )

        return URL

    def files(
            self,
            platform: str,
            product: str,
            acquisition_date: date or str,
            build: str = None,
            pattern: str = None) -> List[str]:
        URL = self.date_URL(platform, product, acquisition_date, build)
        listing = self.get_HTTP_listing(URL, pattern)

        return listing

    def download_URL(
            self,
            URL: str,
            download_location: str = None,
            XML_retries: int = None,
            XML_timeout_seconds: int = None,
            download_retries: int = None,
            download_wait_seconds: int = None,
            enforce_checksum: bool = False) -> str:
        if isdir(download_location):
            filename = join(download_location, posixpath.basename(URL))
        else:
            filename = download_location

        if exists(filename):
            logger.info(f"file already retrieved: {cl.file(filename)}")
            return filename

        # metadata = self.get_metadata(URL)
        metadata_URL = f"{URL}.xml"
        logger.info(f"checking metadata: {cl.URL(metadata_URL)}")

        if isdir(download_location):
            metadata_filename = join(download_location, posixpath.basename(metadata_URL))
        else:
            metadata_filename = f"{download_location}.xml"

        makedirs(abspath(dirname(expanduser(metadata_filename))), exist_ok=True)

        if XML_retries is None:
            XML_retries = XML_RETRIES

        if XML_timeout_seconds is None:
            XML_timeout_seconds = XML_TIMEOUT

        if download_retries is None:
            download_retries = DOWNLOAD_RETRIES

        if download_wait_seconds is None:
            download_wait_seconds = DOWNLOAD_WAIT_SECONDS

        metadata = None

        while XML_retries > 0:
            XML_retries -= 1
            command = f"wget -nc -c --user {self._username} --password {self._password} -O {abspath(expanduser(metadata_filename))} {metadata_URL}"
            # logger.info(command)
            os.system(command)

            if not exists(metadata_filename):
                logger.warning(f"download not found for metadata URL: {metadata_URL}")
                logger.warning(f"waiting {XML_timeout_seconds} for retry")
                sleep(XML_timeout_seconds)
                continue

            XML_metadata_filesize = self.get_local_filesize(metadata_filename)

            if XML_metadata_filesize == 0 and exists(metadata_filename):
                logger.warning(f"removing corrupted zero-size metadata file: {metadata_filename}")

                try:
                    os.remove(metadata_filename)
                except:
                    logger.warning(f"unable to remove zero-size metadata file: {metadata_filename}")

                logger.warning(f"waiting {XML_timeout_seconds} for retry")
                sleep(XML_timeout_seconds)
                continue

            try:
                with open(abspath(expanduser(metadata_filename)), "r") as file:
                    metadata = xmltodict.parse(file.read())
            except Exception as e:
                logger.warning(e)
                logger.warning(f"unable to parse metadata file: {metadata_filename}")
                os.remove(abspath(expanduser(metadata_filename)))
                logger.warning(f"waiting {XML_timeout_seconds} for retry")
                sleep(XML_timeout_seconds)
                continue

        if metadata is None:
            raise DownloadFailed(f"unable to retrieve metadata URL: {metadata_URL}")  # exit code 16

        remote_checksum = str(
            metadata["GranuleMetaDataFile"]["GranuleURMetaData"]["DataFiles"]["DataFileContainer"]["Checksum"])
        checksum_type = str(
            metadata["GranuleMetaDataFile"]["GranuleURMetaData"]["DataFiles"]["DataFileContainer"]["ChecksumType"])
        remote_filesize = int(
            metadata["GranuleMetaDataFile"]["GranuleURMetaData"]["DataFiles"]["DataFileContainer"]["FileSize"])

        logger.info(
            f"metadata retrieved {checksum_type} checksum: {cl.val(remote_checksum)} size: {cl.val(remote_filesize)} URL: {cl.URL(metadata_URL)}")
        makedirs(abspath(dirname(expanduser(filename))), exist_ok=True)
        logger.info(f"downloading {cl.URL(URL)} -> {cl.file(filename)}")

        # Use a temporary file for downloading
        temporary_filename = f"{filename}.download"

        while download_retries > 0:
            download_retries -=1

            try:
                if exists(abspath(expanduser(temporary_filename))):
                    temporary_filesize = self.get_local_filesize(abspath(expanduser(temporary_filename)))

                    if temporary_filesize > remote_filesize:
                        logger.warning(
                            f"removing corrupted file with size {temporary_filesize} greater than remote size {remote_filesize}: {temporary_filename}")
                        remove(temporary_filename)

                    elif temporary_filesize == remote_filesize:
                        # File size matches, now check if we need to validate checksum
                        file_valid = True
                        
                        if enforce_checksum:
                            local_checksum = self.get_local_checksum(abspath(expanduser(temporary_filename)), checksum_type=checksum_type)
                            
                            if local_checksum != remote_checksum:
                                logger.warning(
                                    f"removing corrupted file with local checksum {local_checksum} and remote checksum {remote_checksum}: {temporary_filename}")
                                remove(abspath(expanduser(temporary_filename)))
                                file_valid = False
                            else:
                                logger.info(f"checksum validation passed: {local_checksum}")
                        else:
                            logger.info("skipping checksum validation (enforce_checksum=False)")
                        
                        if file_valid:
                            try:
                                shutil.move(abspath(expanduser(temporary_filename)), abspath(expanduser(filename)))
                            except Exception as e:
                                if exists(filename):
                                    logger.warning(f"unable to move temporary file: {temporary_filename}")
                                    return filename

                                logger.exception(e)
                                raise DownloadFailed(f"unable to move temporary file: {temporary_filename}")

                            return filename
                    else:
                        logger.info(f"resuming incomplete download: {cl.file(temporary_filename)}")

                command = f"wget -nc -c --user {self._username} --password {self._password} -O {abspath(expanduser(temporary_filename))} {URL}"
                # logger.info(command)
                os.system(command)

                if not exists(abspath(expanduser(temporary_filename))):
                    raise ConnectionError(f"unable to download URL: {URL}")

                local_filesize = self.get_local_filesize(abspath(expanduser(temporary_filename)))
                
                # Step 1: Validate file size (always required)
                if local_filesize != remote_filesize:
                    os.remove(abspath(expanduser(temporary_filename)))
                    raise ConnectionError(
                        f"removing corrupted file with incorrect filesize - local: {local_filesize}, remote: {remote_filesize}: {temporary_filename}")
                
                # Step 2: Validate checksum (optional based on enforce_checksum flag)
                if enforce_checksum:
                    local_checksum = self.get_local_checksum(abspath(expanduser(temporary_filename)), checksum_type=checksum_type)
                    
                    if local_checksum != remote_checksum:
                        os.remove(abspath(expanduser(temporary_filename)))
                        raise ConnectionError(
                            f"removing corrupted file with incorrect checksum - local: {local_checksum}, remote: {remote_checksum}: {temporary_filename}")
                    
                    logger.info(f"checksum validation passed: {local_checksum}")
                else:
                    logger.info("skipping checksum validation (enforce_checksum=False)")
                    local_checksum = "skipped"

                # Download successful, rename the temporary file to its proper name
                shutil.move(abspath(expanduser(temporary_filename)), abspath(expanduser(filename)))

                if enforce_checksum:
                    logger.info(
                        f"successful download with filesize {cl.val(local_filesize)} checksum {cl.val(local_checksum)}: {cl.file(filename)}")
                else:
                    logger.info(
                        f"successful download with filesize {cl.val(local_filesize)} (checksum validation skipped): {cl.file(filename)}")

                return filename
            except Exception as e:
                if download_retries == 0:
                    raise e
                else:
                    logger.warning(e)
                    logger.warning(f"waiting {download_wait_seconds} seconds to retry download")
                    sleep(download_wait_seconds)
                    continue



from typing import Union
import json
import logging
import os
import posixpath
from os import makedirs
from os.path import exists, splitext, basename, join, expanduser, abspath, dirname
from typing import List
from zipfile import ZipFile

import numpy as np
from shapely.geometry import Polygon

import colored_logging as cl
from rasters import Raster, RasterGeometry, RasterGrid, Point, MultiPoint
import rasters

import pandas as pd

from .LPDAACDataPool import LPDAACDataPool


DEFAULT_WORKING_DIRECTORY = join("~", "data", "NASADEM")
DEFAULT_DOWNLOAD_DIRECTORY = DEFAULT_WORKING_DIRECTORY

SRTM_FILENAMES_CSV = join(abspath(dirname(__file__)), "filenames.csv")

logger = logging.getLogger(__name__)

class NASADEMGranule:
    def __init__(self, filename: str):
        if not exists(filename):
            raise IOError(f"SRTM file not found: {filename}")

        self.filename = filename
        self._geometry = None

    @property
    def tile(self):
        return splitext(basename(self.filename))[0].split("_")[-1]

    @property
    def hgt_URI(self) -> str:
        return f"zip://{self.filename}!/{self.tile.lower()}.hgt"


    def get_elevation_m(self, geometry: Union[RasterGeometry, Point, MultiPoint] = None) -> Raster:
        URI = self.hgt_URI
        logger.info(f"loading elevation: {cl.URL(URI)}")
        data = Raster.open(URI, geometry=geometry)

        if isinstance(data, Raster):
            data = rasters.where(data == data.nodata, np.nan, data.astype(np.float32))
            data.nodata = np.nan

        return data

    elevation_m = property(get_elevation_m)

    # @property
    # def ocean(self) -> Raster:
    #     return self.elevation_m == 0

    @property
    def geometry(self) -> RasterGrid:
        if self._geometry is None:
            self._geometry = RasterGrid.open(self.hgt_URI)

        return self._geometry

    @property
    def swb_URI(self) -> str:
        return f"zip://{self.filename}!/{self.tile.lower()}.swb"

    def get_swb(self, geometry: Union[RasterGeometry, Point, MultiPoint] = None) -> Raster:
        URI = self.swb_URI
        
        if geometry is None:
            geometry = self.geometry

        filename = self.filename
        member_name = f"{self.tile.lower()}.swb"
        logger.info(f"loading swb: {cl.URL(URI)}")

        with ZipFile(filename, "r") as zip_file:
            with zip_file.open(member_name, "r") as file:
                data = Raster(np.frombuffer(file.read(), dtype=np.int8).reshape(geometry.shape), geometry=geometry)

        data = data != 0

        return data
    
    swb = property(get_swb)


class TileNotAvailable(ValueError):
    pass


class NASADEMConnection(LPDAACDataPool):
    # logger = logging.getLogger(__name__)

    def __init__(
            self,
            username: str = None,
            password: str = None,
            remote: str = None,
            working_directory: str = None,
            download_directory: str = None,
            offline_ok: bool = True):
        super(NASADEMConnection, self).__init__(username=username, password=password, remote=remote, offline_ok=offline_ok)

        if working_directory is None:
            working_directory = DEFAULT_WORKING_DIRECTORY
        self.working_directory = abspath(expanduser(working_directory))
        # logger.info(f"SRTM working directory: {cl.dir(working_directory)}")

        if download_directory is None:
            download_directory = DEFAULT_DOWNLOAD_DIRECTORY
        self.download_directory = abspath(expanduser(download_directory))
        # logger.info(f"SRTM download directory: {cl.dir(download_directory)}")

        self._filenames = None

    def __repr__(self):
        display_dict = {
            "URL": self.remote,
            "download_directory": self.download_directory
        }

        display_string = json.dumps(display_dict, indent=2)

        return display_string

    @property
    def filenames(self) -> List[str]:
        if self._filenames is None:
            self._filenames = list(pd.read_csv(SRTM_FILENAMES_CSV)["filename"])

        return self._filenames

    def tile_URL(self, tile: str) -> str:
        return posixpath.join(
            self.remote,
            "MEASURES",
            "NASADEM_HGT.001",
            "2000.02.11",
            f"NASADEM_HGT_{tile.lower()}.zip"
        )
    
    def tile_filename(self, tile: str) -> str:
        return join(
            self.download_directory,
            f"NASADEM_HGT_{tile.lower()}.zip"
        )

    def tiles_intersecting_bbox(self, lon_min, lat_min, lon_max, lat_max):
        tiles = []

        for lat in range(int(np.floor(lat_min)), int(np.floor(lat_max)) + 1):
            for lon in range(int(np.floor(lon_min)), int(np.floor(lon_max)) + 1):
                tiles.append(f"{'s' if lat < 0 else 'n'}{abs(lat):02d}{'w' if lon < 0 else 'e'}{abs(lon):03d}")

        tiles = sorted(tiles)

        return tiles

    def tiles_intersecting_polygon(self, polygon: Polygon):
        lons, lats = polygon.exterior.xy
        lon_min, lon_max = np.nanmin(lons), np.nanmax(lons)
        lat_min, lat_max = np.nanmin(lats), np.nanmax(lats)

        return self.tiles_intersecting_bbox(lon_min, lat_min, lon_max, lat_max)

    def tiles_intersecting_point(self, point: Point):
        return self.tiles_intersecting_bbox(point.x, point.y, point.x, point.y)

    def tiles_intersecting_multipoint(self, multipoint: MultiPoint):
        return self.tiles_intersecting_bbox(multipoint.xmin, multipoint.ymin, multipoint.xmax, multipoint.ymax)

    def tiles(self, geometry: Union[Polygon, RasterGeometry, Point, MultiPoint]) -> List[str]:
        if isinstance(geometry, Polygon):
            return self.tiles_intersecting_polygon(geometry)
        elif isinstance(geometry, RasterGeometry):
            return self.tiles_intersecting_polygon(geometry.boundary_latlon)
        elif isinstance(geometry, Point):
            return self.tiles_intersecting_point(geometry)
        elif isinstance(geometry, MultiPoint):
            return self.tiles_intersecting_multipoint(geometry)
        else:
            raise ValueError("invalid target geometry")

    def download_tile(self, tile: str, enforce_checksum: bool = False) -> NASADEMGranule:
        filename = self.tile_filename(tile)

        if not exists(filename):
            URL = self.tile_URL(tile)
            filename_base = posixpath.basename(URL)

            if filename_base not in self.filenames:
                raise TileNotAvailable(f"SRTM does not cover tile {tile}")

            directory = self.download_directory
            makedirs(directory, exist_ok=True)
            logger.info(f"acquiring SRTM tile: {cl.val(tile)} URL: {cl.URL(URL)}")
            filename = self.download_URL(URL, directory, enforce_checksum=enforce_checksum)

        granule = NASADEMGranule(filename)

        return granule

    def swb(self, geometry: Union[RasterGeometry, Point, MultiPoint]):
        """
        digital elevation model surface water body from the Shuttle Rader Topography Mission (SRTM)
        :param geometry: target geometry
        :return: raster of elevation for RasterGeometry, values for Point/MultiPoint
        """
        if isinstance(geometry, (Point, MultiPoint)):
            return self._extract_point_values(geometry, 'swb')
        
        result = Raster(np.full(geometry.shape, np.nan, dtype=np.float32), geometry=geometry)
        tiles = self.tiles(geometry)

        for tile in tiles:
            # logger.info(f"acquiring SRTM tile {tile}")
            try:
                granule = self.download_tile(tile)
            except TileNotAvailable as e:
                logger.warning(e)
                continue
            
            image = granule.swb.astype(np.float32)
            image_projected = image.to_geometry(geometry)
            result = rasters.where(np.isnan(result), image_projected, result)

        result = rasters.where(np.isnan(result), 1, result)
        result = result.astype(bool)

        return result

    def elevation_m(self, geometry: Union[RasterGeometry, Point, MultiPoint]):
        """
        digital elevation model (elevation_km) in meters from the Shuttle Rader Topography Mission (SRTM)
        :param geometry: target geometry
        :return: raster of elevation for RasterGeometry, values for Point/MultiPoint
        """
        if isinstance(geometry, MultiPoint):
            return self._extract_point_values(geometry, 'elevation_m')
        
        result = None
        tiles = self.tiles(geometry)

        for tile in tiles:
            # logger.info(f"acquiring SRTM tile {tile}")
            try:
                granule = self.download_tile(tile)
            except TileNotAvailable as e:
                logger.warning(e)
                continue

            elevation_m = granule.get_elevation_m(geometry=geometry)
            print(elevation_m)
            # elevation_projected = elevation_m.to_geometry(geometry)

            if result is None:
                result = elevation_m
            elif isinstance(result, Raster):
                result = rasters.where(np.isnan(result), elevation_m, result)

        return result

    def elevation_km(self, geometry: Union[RasterGeometry, Point, MultiPoint]):
        """
        digital elevation model (elevation_km) in kilometers from the Shuttle Rader Topography Mission (SRTM)
        :param geometry: target geometry
        :return: raster of elevation for RasterGeometry, values for Point/MultiPoint
        """
        return self.elevation_m(geometry=geometry) / 1000

    def _extract_point_values(self, geometry: Union[Point, MultiPoint], data_type: str):
        """
        Extract values at point locations from NASADEM data
        :param geometry: Point or MultiPoint geometry
        :param data_type: 'elevation_m' or 'swb'
        :return: single value for Point, array of values for MultiPoint
        """
        tiles = self.tiles(geometry)
        
        if isinstance(geometry, Point):
            points = [geometry]
        else:
            points = [Point(geom.x, geom.y) for geom in geometry.geoms]
        
        values = []
        
        for point in points:
            point_value = np.nan
            
            for tile in tiles:
                try:
                    granule = self.download_tile(tile)
                except TileNotAvailable as e:
                    logger.warning(e)
                    continue
                
                # Get the appropriate data layer
                if data_type == 'elevation_m':
                    data_layer = granule.elevation_m
                elif data_type == 'swb':
                    data_layer = granule.swb.astype(np.float32)
                else:
                    raise ValueError(f"Unknown data type: {data_type}")
                
                # Check if point is within this tile's bounds
                tile_geometry = data_layer.geometry
                if (tile_geometry.x_min <= point.x <= tile_geometry.x_max and 
                    tile_geometry.y_min <= point.y <= tile_geometry.y_max):
                    
                    # Extract value at point location
                    try:
                        point_raster = data_layer.to_point(point)
                        if hasattr(point_raster, 'values') and len(point_raster.values) > 0:
                            point_value = point_raster.values[0]
                        else:
                            point_value = np.nan
                        if not np.isnan(point_value):
                            break  # Found valid value, stop searching tiles
                    except Exception as e:
                        logger.warning(f"Failed to sample point ({point.x}, {point.y}) from tile {tile}: {e}")
                        continue
            
            values.append(point_value)
        
        # Apply post-processing for swb
        if data_type == 'swb':
            values = [bool(v != 0) if not np.isnan(v) else False for v in values]
        
        # Return single value for Point, array for MultiPoint
        if isinstance(geometry, Point):
            return values[0]
        else:
            return np.array(values)

NASADEM = NASADEMConnection()
